import Foundation
import Flutter
import Ably
import UIKit

class AblyFlutterReader: FlutterStandardReader {
    override init(data: Data) {
        super.init(data: data)
    }

    override func readValue(ofType type: UInt8) -> Any? {
        let value = readValue(ofType: type)
        if let valueString = value as? String {
            switch type {
            case CodecType.logLevel.rawValue: return ARTLogLevel.from(logLevel: valueString)
            default: return super.readValue(ofType: type)
            }
        } else if let valueDictionary = value as? Dictionary<String, Any> {
            switch type {
            case CodecType.ablyMessage.rawValue: return decodeMessage(dictionary: valueDictionary)
            case CodecType.ablyEventMessage.rawValue: return decodeEventMessage(dictionary: valueDictionary)
            case CodecType.clientOptions.rawValue: return AblyFlutterClientOptions.from(dictionary: valueDictionary)
            case CodecType.messageExtras.rawValue:
            case CodecType.message.rawValue:
            case CodecType.tokenDetails.rawValue:
            case CodecType.tokenRequest.rawValue:
            case CodecType.restChannelOptions.rawValue:
            case CodecType.realtimeChannelOptions.rawValue:
            case CodecType.restHistoryParams.rawValue:
            case CodecType.realtimeHistoryParams.rawValue:
            case CodecType.restPresenceParams.rawValue:
            case CodecType.realtimePresenceParams.rawValue:
            case CodecType.messageData.rawValue:
            default: return super.readValue(ofType: type)
            }
        }

    }

    private func decodeMessage(dictionary: Dictionary<String, Any?>) -> AblyFlutterMessage {
        let type = dictionary[TxAblyMessage_type] as! UInt8?
        let handle = dictionary[TxAblyMessage_registrationHandle] as! Int
        let message = dictionary[TxAblyMessage_message] as! NSDictionary
        if type != nil {
            let decodedMessage = readValue(ofType: type!)!
            return AblyFlutterMessage(message: decodedMessage, handle: NSNumber(value: handle))
        }
        return AblyFlutterMessage(message: message, handle: NSNumber(value: handle))
    }

    private func decodeEventMessage(dictionary: Dictionary<String, Any?>) -> EventMessage {
        let type = dictionary[TxAblyMessage_type] as! UInt8?
        let eventName = dictionary[TxAblyEventMessage_eventName] as! String
        let handle = dictionary[TxAblyEventMessage_registrationHandle] as! Int
        let message = dictionary[TxAblyMessage_message] as! NSDictionary
        if type != nil {
            let decodedMessage = readValue(ofType: type!)!
            return EventMessage(eventName: eventName, message: decodedMessage, handle: handle)
        }
        return EventMessage(eventName: eventName, message: message, handle: handle)
    }
}

extension AblyFlutterClientOptions {
    static func from(dictionary: Dictionary<String, Any>) -> AblyFlutterClientOptions {
        let clientOptions = ARTClientOptions()
        // TODO authUrl might not work, since this value is definitely not a Url! (Carried over from original Objective-C code)
        if let authUrl = dictionary[TxClientOptions_authUrl] as? String {
            clientOptions.authUrl = URL(string: authUrl)
        }
        clientOptions.authMethod = dictionary[TxClientOptions_authMethod] as? String
        clientOptions.key = dictionary[TxClientOptions_key] as? String
        clientOptions.tokenDetails = ARTTokenDetails.from(dictionary: dictionary[TxClientOptions_tokenDetails] as! Dictionary<String, Any>)
        clientOptions.authHeaders = dictionary[TxClientOptions_authHeaders] as! Dictionary<String, String>
        if let authParamsDictionary = dictionary[TxClientOptions_authParams] as! Dictionary<String, String>? {
            clientOptions.authParams = authParamsDictionary.map { key, value in URLQueryItem(name: key, value: value) }
        }
        if let queryTime = dictionary[TxClientOptions_queryTime] as? Bool { clientOptions.queryTime = queryTime }
        clientOptions.clientId = dictionary[TxClientOptions_clientId] as? String
        clientOptions.tls = dictionary[TxClientOptions_tls] as? Bool
        clientOptions.restHost = dictionary[TxClientOptions_restHost] as? String
        clientOptions.realtimeHost = dictionary[TxClientOptions_realtimeHost] as? String
        clientOptions.autoConnect = dictionary[TxClientOptions_autoConnect] as? Bool
        clientOptions.useBinaryProtocol = dictionary[TxClientOptions_useBinaryProtocol] as? Bool
        clientOptions.queueMessages = dictionary[TxClientOptions_queueMessages] as? Bool
        clientOptions.echoMessages = dictionary[TxClientOptions_echoMessages] as? Bool
        clientOptions.recover = dictionary[TxClientOptions_recover] as? String
        clientOptions.environment = dictionary[TxClientOptions_environment] as? String
        clientOptions.idempotentRestPublishing = dictionary[TxClientOptions_idempotentRestPublishing] as? Bool
        clientOptions.fallbackHosts = dictionary[TxClientOptions_fallbackHosts] as [String]?
        clientOptions.defaultTokenParams = ARTTokenParams.from(dictionary: dictionary[TxClientOptions_defaultTokenParams] as! Dictionary<String, Any>)

        clientOptions.logLevel = ARTLogLevel.from(logLevel: dictionary[TxClientOptions_logLevel] as String)
        clientOptions.addAgent("ably-flutter", version: FLUTTER_PACKAGE_PLUGIN_VERSION)

        let hasAuthCallback = dictionary[TxClientOptions_hasAuthCallback] as! Bool
        return AblyFlutterClientOptions(clientOptions: clientOptions, hasAuthCallback: hasAuthCallback)
    }
}

extension ARTLogLevel {
    static func from(logLevel: String) -> ARTLogLevel {
        switch (logLevel) {
        case TxLogLevelEnum_verbose: return .verbose
        case TxLogLevelEnum_debug: return .debug
        case TxLogLevelEnum_info: return .info
        case TxLogLevelEnum_warn: return .warn
        case TxLogLevelEnum_error: return .error
        case TxLogLevelEnum_none: return .none
        default: return .warn
        }
    }
}

extension ARTTokenParams {
    static func from(dictionary: Dictionary<String, Any>) -> ARTTokenParams {
        let clientId = dictionary[TxTokenParams_clientId] as? String
        let nonce = dictionary[TxTokenParams_nonce] as? String
        let tokenParams = ARTTokenParams(clientId: clientId, nonce: nonce)

        tokenParams.capability = dictionary[TxTokenParams_capability] as? String
        if let ttl = dictionary[TxTokenParams_ttl] {
            tokenParams.ttl = NSNumber(ttl as! Double)
        }

        if let timestamp = dictionary[TxTokenParams_timestamp] {
            // TODO document in the variable, that TxTokenParams_timestamp is a Double. e.g. TxTokenParams_timestampInSecondsAsDouble
            tokenParams.timestamp = decodeTimestamp(inMilliseconds: timestamp as! Double)
        }
        return tokenParams
    }
}

func decodeTimestamp(inMilliseconds: Double) -> Date {
    Date(timeIntervalSince1970: inMilliseconds / 1000)
}

func decodeTimestamp(inSeconds: Double) -> Date {
    Date(timeIntervalSince1970: inSeconds)
}

extension ARTTokenDetails {
    static func from(dictionary: Dictionary<String, Any>) -> ARTTokenDetails {
        let token = dictionary[TxTokenDetails_token] as? String
        // TODO confirm that these are NSDates!
        let expires = dictionary[TxTokenDetails_expires] as? NSDate
        let issued = dictionary[TxTokenDetails_issued] as? NSDate
        let capability = dictionary[TxTokenDetails_capability] as? String
        let clientId = dictionary[TxTokenDetails_clientId] as? String

        return ARTTokenDetails(token: token, expires: expires, issued: issued, capability: capability, clientId: clientId)
    }
}
